@@ -323,9 +323,10 @@ _dbus_win_error_from_last_error (void)
     case ERROR_FILE_NOT_FOUND:
     case ERROR_PATH_NOT_FOUND:
       return DBUS_ERROR_FILE_NOT_FOUND;
+
+    default:
+      return DBUS_ERROR_FAILED;
     }
-  
-  return DBUS_ERROR_FAILED;
 }
 
 
@@ -2226,6 +2227,35 @@ _dbus_disable_sigpipe (void)
 {
 }
 
+/**
+ * Creates a directory. Unlike _dbus_ensure_directory(), this only succeeds
+ * if the directory is genuinely newly-created.
+ *
+ * @param filename directory filename
+ * @param error initialized error object
+ * @returns #TRUE on success
+ */
+dbus_bool_t
+_dbus_create_directory (const DBusString *filename,
+                        DBusError        *error)
+{
+  const char *filename_c;
+
+  _DBUS_ASSERT_ERROR_IS_CLEAR (error);
+
+  filename_c = _dbus_string_get_const_data (filename);
+
+  if (!CreateDirectoryA (filename_c, NULL))
+    {
+      dbus_set_error (error, DBUS_ERROR_FAILED,
+                      "Failed to create directory %s: %s\n",
+                      filename_c, _dbus_strerror_from_errno ());
+      return FALSE;
+    }
+  else
+    return TRUE;
+}
+
 /**
  * Creates a directory; succeeds if the directory
  * is created or already existed.
@@ -2235,7 +2265,7 @@ _dbus_disable_sigpipe (void)
  * @returns #TRUE on success
  */
 dbus_bool_t
-_dbus_create_directory (const DBusString *filename,
+_dbus_ensure_directory (const DBusString *filename,
                         DBusError        *error)
 {
   const char *filename_c;
@@ -2687,9 +2717,9 @@ HANDLE _dbus_global_lock (const char *mutexname)
        case WAIT_FAILED:
        case WAIT_TIMEOUT:
                return 0;
+       default:
+               return mutex;
      }
-
-   return mutex;
 }
 
 static
@@ -3524,12 +3554,15 @@ _dbus_strerror (int error_number)
 
     case WSASYSCALLFAILURE:
       return "System call failure";
-    }
-  msg = strerror (error_number);
-  if (msg == NULL)
-    msg = "unknown";
 
-  return msg;
+    default:
+      msg = strerror (error_number);
+
+      if (msg == NULL)
+        msg = "unknown";
+
+      return msg;
+    }
 #endif //DBUS_WINCE
 }
 
@@ -3669,9 +3702,6 @@ _dbus_init_system_log (const char   *tag,
  * @param severity a severity value
  * @param msg a printf-style format string
  * @param args arguments for the format string
- *
- * If the FATAL severity is given, this function will terminate the program
- * with an error code.
  */
 void
 _dbus_logv (DBusSystemLogSeverity  severity,
@@ -3686,7 +3716,8 @@ _dbus_logv (DBusSystemLogSeverity  severity,
      case DBUS_SYSTEM_LOG_INFO: s = "info"; break;
      case DBUS_SYSTEM_LOG_WARNING: s = "warning"; break;
      case DBUS_SYSTEM_LOG_SECURITY: s = "security"; break;
-     case DBUS_SYSTEM_LOG_FATAL: s = "fatal"; break;
+     case DBUS_SYSTEM_LOG_ERROR: s = "error"; break;
+     default: _dbus_assert_not_reached ("invalid log severity");
    }
 
   if (log_flags & DBUS_LOG_FLAGS_SYSTEM_LOG)
@@ -3709,9 +3740,6 @@ _dbus_logv (DBusSystemLogSeverity  severity,
       fprintf (stderr, "\n");
       va_end (tmp);
     }
-
-  if (severity == DBUS_SYSTEM_LOG_FATAL)
-    exit (1);
 }
 
 /** @} end of sysdeps-win */
